package net.lemnik.eodsql;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;

import java.util.Arrays;

/**
 *<p>
 * An implementation of {@link DataSetCache} that uses an array of
 * {@link java.lang.ref.WeakReference}'s to store the row values.
 * This is the default cache implementation.
 *</p><p>
 * While fairly quick, an <code>ArrayDataSetCache</code> does hold
 * references for as long as you do. If you do not need a cache for
 * identity equality an {@code ArrayDataSetCache} is probably overkill.
 * For large <code>DataSet</code>'s it is advised that you either look
 * at a {@link NullDataSetCache}, or into using a {@link DataIterator}
 * instead of a {@code DataSet}.
 *</p>
 *
 * @param <T> the row data-type to cache
 * @author Jason Morris
 * @see NullDataSetCache
 * @see DataIterator
 */
public class ArrayDataSetCache<T> implements DataSetCache<T> {
    private static final int DEFAULT_SIZE = 10;

    private ReferenceQueue<T> queue = new ReferenceQueue<T>();

    @SuppressWarnings("unchecked")
    private WeakReference[] cache = new WeakReference[DEFAULT_SIZE];

    /**
     * Default constructor for {@code ArrayDataSetCache}.
     * This will create an internal {@link WeakReference} cache of a
     * preset size, ready for objects to be cached.
     */
    public ArrayDataSetCache() {
    }

    @SuppressWarnings("unchecked")
    private void ensureCapacity(final int newCapacity) {
        final int len = cache.length;
        if(len < newCapacity) {
            final WeakReference<?>[] newCache = new WeakReference<?>[Math.max((len * 3) / 2 + 1, newCapacity)];
            System.arraycopy(cache, 0, newCache, 0, len);
            cache = newCache;        
        }
    }

    @SuppressWarnings("unchecked")
    private void removeOldEntries() {
        IndexedWeakReference ref = null;

        while((ref = (IndexedWeakReference)queue.poll()) != null) {
            // allow the collector to GC the reference object
            cache[ref.getIndex()] = null;
        }
    }

    public void init(final DataSet<T> dataSet) {
    }

    public boolean isCached(final int row) {
        return (row < cache.length &&
                cache[row] != null &&
                cache[row].get() != null);
    }

    public T getObject(final int row) {
        removeOldEntries();

        if(isCached(row)) {
            @SuppressWarnings("unchecked")
            final T value = (T)cache[row].get();
            return value;
        }

        return null;
    }

    public void setObject(final int row, final T object) {
        removeOldEntries();
        ensureCapacity(row + 1);
        cache[row] = new IndexedWeakReference<T>(row, object, queue);
    }

    public void destroy() {
        cache = null;
        queue = null;
    }

    private static class IndexedWeakReference<T> extends WeakReference<T> {
        private final int index;

        IndexedWeakReference(
                final int index,
                final T obj,
                final ReferenceQueue<T> queue) {
            
            super(obj, queue);
            this.index = index;
        }

        public int getIndex() {
            return index;
        }

    }

}
