package net.lemnik.eodsql;

import java.lang.reflect.InvocationTargetException;

import java.sql.Connection;

import java.util.Collections;
import java.util.Iterator;

import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import javax.sql.DataSource;

import net.lemnik.eodsql.impl.DefaultQueryFactory;
import net.lemnik.eodsql.spi.QueryFactory;

/**
 * The {@code QueryFactoryChain} is responsible for creating the
 * implementations of {@link BaseQuery} object. The {@code QueryFactoryChain}
 * holds a list of the different {@code QueryFactory} implementations
 * available. The {@code QueryFactoryChain} has special reentrant behaviour
 * to cater for {@code QueryFactory} implementations that create delegating
 * (proxy) implementations. If a {@code QueryFactory} currently under
 * construction creates a {@code BaseQuery} implementation, the
 * {@code QueryFactoryChain} will only look at {@code QueryFactories} that
 * appear after the current one.
 *
 * @author Jason Morris
 */
class QueryFactoryChain {

    private final List<QueryFactory> factories =
            new CopyOnWriteArrayList<QueryFactory>();

    private final ThreadLocal<CreateContext> chain =
            new ThreadLocal<CreateContext>() {

        @Override
        protected CreateContext initialValue() {
            return new CreateContext(factories.iterator());
        }

    };

    QueryFactoryChain() {
        add(new DefaultQueryFactory());

        try {
            @SuppressWarnings("unchecked")
            final Iterable<QueryFactory> loader = (Iterable<QueryFactory>)Class.
                    forName("java.util.ServiceLoader").
                    getMethod("load", Class.class).
                    invoke(null, QueryFactory.class);

            for(final QueryFactory factory : loader) {
                add(factory);
            }
        } catch(final NoClassDefFoundError e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final IllegalArgumentException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final SecurityException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final IllegalAccessException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final InvocationTargetException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final NoSuchMethodException e) {
            // ignore this... we are simply not running 1.6 or higher
        } catch(final ClassNotFoundException e) {
            // ignore this... we are simply not running 1.6 or higher
        }
    }

    <T extends BaseQuery> T create(
            final DataSource ds,
            final Class<T> query,
            final ClassLoader loader)
            throws InvalidQueryException {

        T queryImpl = null;

        final CreateContext context = chain.get();
        context.incEntries();

        try {
            final Iterator<QueryFactory> iterator = context.iterator;

            while(iterator.hasNext()) {
                final QueryFactory factory = iterator.next();

                if(factory.canConstruct(query)) {
                    queryImpl = factory.construct(ds, query, loader);
                    break;
                }
            }
        } finally {
            // cleanup after ourselves
            if(context.decEntries() == 0) {
                chain.remove();
            }
        }

        return queryImpl;
    }

    <T extends BaseQuery> T create(
            final Connection connection,
            final Class<T> query,
            final ClassLoader loader)
            throws InvalidQueryException {

        final CreateContext context = chain.get();
        context.incEntries();

        T queryImpl = null;

        try {
            final Iterator<QueryFactory> iterator = context.iterator;

            while(iterator.hasNext()) {
                final QueryFactory factory = iterator.next();

                if(factory.canConstruct(query)) {
                    queryImpl = factory.construct(connection, query, loader);
                    break;
                }
            }
        } finally {
            // cleanup after ourselves
            if(context.decEntries() == 0) {
                chain.remove();
            }
        }

        return queryImpl;
    }

    void add(final QueryFactory factory) {
        // we add factories to the begining of the list
        // this means the DefaultQueryFactory will always appear last
        factories.add(0, factory);
    }

    void remove(final QueryFactory factory) {
        // remove /all/ instances of the given factory
        factories.remove(Collections.singleton(factory));
    }

    private static class CreateContext {

        private final Iterator<QueryFactory> iterator;

        private int entries = 0;

        public CreateContext(final Iterator<QueryFactory> iterator) {
            assert iterator != null;

            this.iterator = iterator;
        }

        public int incEntries() {
            return ++entries;
        }

        public int decEntries() {
            return --entries;
        }

    }
}
